/*T
   Concepts: KSP^solving a system of linear equations
   Concepts: KSP^Laplacian, 2d
   Processors: n
T*/

/*
Added at the request of Marc Garbey.

Inhomogeneous Laplacian in 2D. Modeled by the partial differential equation

   div \rho grad u = f,  0 < x,y < 1,

with forcing function

   f = e^{-x^2/\nu} e^{-y^2/\nu}

with Dirichlet boundary conditions

   u = f(x,y) for x = 0, x = 1, y = 0, y = 1

or pure Neumman boundary conditions

This uses multigrid to solve the linear system
*/

static char help[] = "Solves 2D inhomogeneous Laplacian using multigrid.\n\n";

#include <petscdmda.h>
#include <petscksp.h>
#include <petscpcmg.h>
#include <petscdmmg.h>

extern PetscErrorCode ComputeMatrix(DMMG,Mat,Mat);
extern PetscErrorCode ComputeRHS(DMMG,Vec);
extern PetscErrorCode VecView_VTK(Vec, const char [], const char [], DM);
extern PetscErrorCode CheckSolution(DM, Vec, Vec);
extern PetscErrorCode ComputeRHStest(DMMG,Vec);



typedef enum {DIRICHLET, NEUMANN} BCType;

typedef struct {
  PetscScalar   rho;
  PetscScalar   nu;
  BCType        bcType;
} UserContext;

const double dxValue = 0.025;
const double dyValue = 0.025;
const int imax = 121;
const int jmax = 111;

#undef __FUNCT__
#define __FUNCT__ "main"
int main(int argc,char **argv)
{
  DMMG           *dmmg;
  DM             da;
  UserContext    user;
  PetscReal      norm;
  const char     *bcTypes[2] = {"dirichlet","neumann"};
  PetscErrorCode ierr;
  PetscInt       l,bc;

  PetscInitialize(&argc,&argv,(char *)0,help);

  ierr = DMMGCreate(PETSC_COMM_WORLD,3,PETSC_NULL,&dmmg);CHKERRQ(ierr);
  ierr = DMDACreate2d(PETSC_COMM_WORLD, DMDA_BOUNDARY_NONE, DMDA_BOUNDARY_NONE,DMDA_STENCIL_STAR,-imax,-jmax,PETSC_DECIDE,PETSC_DECIDE,1,1,0,0,&da);CHKERRQ(ierr); 
  ierr = DMDASetUniformCoordinates(da, 0.0, dxValue*(imax-1)*4., 0.0, dyValue*(jmax-1)*4., PETSC_NULL, PETSC_NULL);CHKERRQ(ierr); 
  ierr = DMMGSetDM(dmmg,(DM)da);CHKERRQ(ierr);
//  ierr = DMDestroy(&da);CHKERRQ(ierr);
  for (l = 0; l < DMMGGetLevels(dmmg); l++) {
    ierr = DMMGSetUser(dmmg,l,&user);CHKERRQ(ierr);
  }
/*
  ierr = PetscOptionsBegin(PETSC_COMM_WORLD, "", "Options for the inhomogeneous Poisson equation", "DMMG");
    user.rho    = 1.0;
    ierr        = PetscOptionsScalar("-rho", "The conductivity", "ex29.c", user.rho, &user.rho, PETSC_NULL);CHKERRQ(ierr);
    user.nu     = 0.1;
    ierr        = PetscOptionsScalar("-nu", "The width of the Gaussian source", "ex29.c", user.nu, &user.nu, PETSC_NULL);CHKERRQ(ierr);
    bc          = (PetscInt)DIRICHLET;
    ierr        = PetscOptionsEList("-bc_type","Type of boundary condition","ex29.c",bcTypes,2,bcTypes[0],&bc,PETSC_NULL);CHKERRQ(ierr);
    user.bcType = (BCType)bc;
  ierr = PetscOptionsEnd();
*/

  // ---- First ----
  ierr = DMMGSetKSP(dmmg,ComputeRHS,ComputeMatrix);CHKERRQ(ierr);
  if (user.bcType == NEUMANN) {
    ierr = DMMGSetNullSpace(dmmg,PETSC_TRUE,0,PETSC_NULL);CHKERRQ(ierr);
  }

  ierr = DMMGSolve(dmmg);CHKERRQ(ierr);

  ierr = MatMult(DMMGGetJ(dmmg),DMMGGetx(dmmg),DMMGGetr(dmmg));CHKERRQ(ierr);
  ierr = VecAXPY(DMMGGetr(dmmg),-1.0,DMMGGetRHS(dmmg));CHKERRQ(ierr);
  ierr = VecNorm(DMMGGetr(dmmg),NORM_2,&norm);CHKERRQ(ierr);
  /* ierr = PetscPrintf(PETSC_COMM_WORLD,"Residual norm %G\n",norm);CHKERRQ(ierr); */
  ierr = VecAssemblyBegin(DMMGGetx(dmmg));CHKERRQ(ierr);
  ierr = VecAssemblyEnd(DMMGGetx(dmmg));CHKERRQ(ierr);
  ierr = CheckSolution(DMMGGetDM(dmmg), DMMGGetx(dmmg), DMMGGetRHS(dmmg));
  // ==== First ====

  // ---- Second ----
  ierr = DMMGSetKSP(dmmg,ComputeRHStest,ComputeMatrix);CHKERRQ(ierr);
  if (user.bcType == NEUMANN) {
    ierr = DMMGSetNullSpace(dmmg,PETSC_TRUE,0,PETSC_NULL);CHKERRQ(ierr);
  }

  ierr = DMMGSolve(dmmg);CHKERRQ(ierr);

  ierr = MatMult(DMMGGetJ(dmmg),DMMGGetx(dmmg),DMMGGetr(dmmg));CHKERRQ(ierr);
  ierr = VecAXPY(DMMGGetr(dmmg),-1.0,DMMGGetRHS(dmmg));CHKERRQ(ierr);
  ierr = VecNorm(DMMGGetr(dmmg),NORM_2,&norm);CHKERRQ(ierr);
  /* ierr = PetscPrintf(PETSC_COMM_WORLD,"Residual norm %G\n",norm);CHKERRQ(ierr); */
  ierr = VecAssemblyBegin(DMMGGetx(dmmg));CHKERRQ(ierr);
  ierr = VecAssemblyEnd(DMMGGetx(dmmg));CHKERRQ(ierr);
//  ierr = CheckSolution(DMMGGetDM(dmmg), DMMGGetx(dmmg), DMMGGetRHS(dmmg));
  // ==== Second ====



//  ierr = VecView_VTK(DMMGGetRHS(dmmg), "rhs.vtk", bcTypes[user.bcType], DMMGGetDM(dmmg));CHKERRQ(ierr);
//  ierr = VecView_VTK(DMMGGetx(dmmg), "solution.vtk", bcTypes[user.bcType], DMMGGetDM(dmmg));CHKERRQ(ierr);

  ierr = DMDestroy(&da);CHKERRQ(ierr);
  ierr = DMMGDestroy(dmmg);CHKERRQ(ierr);
  ierr = PetscFinalize();

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "CheckSolution"
PetscErrorCode CheckSolution(DM da, Vec x, Vec rhso) {
  FILE *fp;
  int i, j, k = 0, maxi, maxj;
  double _p[600][600], _newp[600][600], rhs[600][600];
  double maxerr = 1.;
  PetscInt ierr,mx,my,mm,nn,xw,yw,xs,ys;
  PetscScalar **SolArray, **SolRHS;

  fp = fopen("AlanCheck.plt", "w");

  ierr = DMDAGetInfo(da, 0, &mx, &my, 0, &mm , &nn, 0,0,0,0,0,0,0);CHKERRQ(ierr);
  ierr = DMDAGetCorners(da,&xs,&ys,0,&xw,&yw,0);CHKERRQ(ierr);
  
  ierr = DMDAVecGetArray(da, x, &SolArray);
  ierr = DMDAVecGetArray(da, rhso, &SolRHS); 
  PetscFunctionBegin;


  for(j = ys; j < ys+yw; j++) {
    for(i = xs; i < xs+xw; i++) {
      _p[j][i] =  SolArray[j][i];
//      _p[j][i] = 1.0;
    }
  }
  if(mx != xs+xw) printf("%d != %d!!!!\n", mx, xs+xw);

  while(k < 500000 && maxerr > 1.0e-10) {
/*
    for(j = ys; j < ys+yw; j++) {
      for(i = xs; i < xs+xw; i++) {
	if(fabs(_p[j][i]) > maxerr) {
	  maxerr = _p[j][i];
	  maxi = i;
	  maxj = j;
	}
      }
    }
*/
    maxerr = 0.0;

    for(j = ys; j < ys+yw; j++) {
      for(i = xs; i < xs+xw; i++) {
        rhs[j][i] = (j >= 0.5*yw && j <= 0.55*yw && i >= 0.5*xw && i <= 0.6*xw)?j:0.0;
//        if(rhs[j][i] != SolRHS[j][i]) printf("%f  !=  %f!!!!\n", rhs[j][i], SolRHS[j][i]);
	if(i == xs && i != xs+xw-1 && j != ys && j != ys+yw-1) _newp[j][i] = (_p[j+1][i] + _p[j-1][i] + _p[j][i+1] - rhs[j][i]*dxValue*dyValue)/3.;
	else if(i != xs && i == xs+xw-1 && j != ys && j != ys+yw-1) _newp[j][i] = 0.;
	else if(i != xs && i != xs+xw-1 && j == ys && j != ys+yw-1) _newp[j][i] = (_p[j+1][i] + _p[j][i+1] + _p[j][i-1] - rhs[j][i]*dxValue*dyValue)/3.;
	else if(i != xs && i != xs+xw-1 && j != ys && j == ys+yw-1) _newp[j][i] = (_p[j][i-1] + _p[j-1][i] + _p[j][i+1] - rhs[j][i]*dxValue*dyValue)/3.;
	else if(i == xs && i != xs+xw-1 && j == ys && j != ys+yw-1) _newp[j][i] = (_p[j+1][i] + _p[j][i+1]  - rhs[j][i]*dxValue*dyValue)/2.;
	else if(i == xs && i != xs+xw-1 && j != ys && j == ys+yw-1) _newp[j][i] = (_p[j-1][i] + _p[j][i+1]  - rhs[j][i]*dxValue*dyValue)/2.;
	else if(i != xs && i == xs+xw-1 && j != ys && j == ys+yw-1) _newp[j][i] = (_p[j-1][i] + _p[j][i-1]  - rhs[j][i]*dxValue*dyValue)/3.;
	else if(i != xs && i == xs+xw-1 && j == ys && j != ys+yw-1) _newp[j][i] = (_p[j+1][i] + _p[j][i-1]  - rhs[j][i]*dxValue*dyValue)/3.;
	else _newp[j][i] = (_p[j+1][i] + _p[j-1][i] + _p[j][i+1] + _p[j][i-1] - rhs[j][i]*dxValue*dyValue)/4.;
/*
        if(i == xs && i != xs+xw-1 && j != ys && j != ys+yw-1) _newp[j][i] = (_p[j+1][i] + _p[j-1][i] + _p[j][i+1] + _p[j][i] - rhs[j][i]*dxValue*dyValue)/4.;
        else if(i != xs && i == xs+xw-1 && j != ys && j != ys+yw-1) _newp[j][i] = 0.;
        else if(i != xs && i != xs+xw-1 && j == ys && j != ys+yw-1) _newp[j][i] = (_p[j+1][i] + _p[j][i+1] + _p[j][i-1] + _p[j][i] - rhs[j][i]*dxValue*dyValue)/4.;
        else if(i != xs && i != xs+xw-1 && j != ys && j == ys+yw-1) _newp[j][i] = (_p[j][i-1] + _p[j-1][i] + _p[j][i+1] + _p[j][i] - rhs[j][i]*dxValue*dyValue)/4.;
        else if(i == xs && i != xs+xw-1 && j == ys && j != ys+yw-1) _newp[j][i] = (_p[j+1][i] + _p[j][i+1] + 2.*_p[j][i] - rhs[j][i]*dxValue*dyValue)/4.;
        else if(i == xs && i != xs+xw-1 && j != ys && j == ys+yw-1) _newp[j][i] = (_p[j-1][i] + _p[j][i+1] + 2.*_p[j][i] - rhs[j][i]*dxValue*dyValue)/4.;
        else if(i != xs && i == xs+xw-1 && j != ys && j == ys+yw-1) _newp[j][i] = 0.;
        else if(i != xs && i == xs+xw-1 && j == ys && j != ys+yw-1) _newp[j][i] = 0.;
        else _newp[j][i] = (_p[j+1][i] + _p[j-1][i] + _p[j][i+1] + _p[j][i-1] - rhs[j][i]*dxValue*dyValue)/4.;
  */      
        if(_p[j][i] != 0 && i!= xs+xw-1) {
          maxerr += pow((_newp[j][i]-_p[j][i])/_p[j][i],2.0);
//          if(pow((_newp[j][i]-_p[j][i])/_p[j][i],2.0) > 0.1) printf("%d (%d, %d) ==> (new = %e, old = %e)\n", xs+xw-1, i, j, _newp[j][i], _p[j][i]);
        }
      }
    }
    maxerr = sqrt(maxerr);
    printf("Inside Check: Iteration%d, Max Value = %e \n", k, maxerr);
//    if(maxerr > 0.00001) maxerr = fabs((_newp[maxj][maxi]-_p[maxj][maxi])/_p[maxj][maxi]);
//    printf("Max Error = %e  \n", maxerr);


    for(j = ys; j < ys+yw; j++) {
      for(i = xs; i < xs+xw; i++) {
	_p[j][i] = _newp[j][i];
      }
    }

    k++;
  }
  printf("Inside Iteration Times: %d\n", k);

  fprintf(fp, "VARIABLES = \"X\", \"Y\", \"U\"\n");
  fprintf(fp, "ZONE I = %d, J= %d, F=POint\n", xw, yw);
  for(j = ys; j < ys+yw; j++) {
    for(i = xs; i < xs+xw; i++) {
      fprintf(fp, "%f  %f  %f  \n", dxValue*i, dyValue*j, _p[j][i]);
    }
  }
  fclose(fp);


  FILE *fppl1, *fppl2, *fppl3;
  fppl1 = fopen("PrsOLine1.dat", "w");
  fppl2 = fopen("PrsOLine2.dat", "w");
  fppl3 = fopen("PrsOLine3.dat", "w");
  for(i = xs; i <= xs+xw-1; i++) {
    for(j = ys; j <= ys+yw-1; j++) {
      if(i == 10) fprintf(fppl1, "%f  %f  \n", dyValue*j, _p[j][i]);
      if(i == (int)0.5*xw) fprintf(fppl2, "%f  %f  \n", dyValue*j, _p[j][i]);
      if(i == (int)0.75*xw) fprintf(fppl3, "%f  %f  \n", dyValue*j, _p[j][i]);
    }
  }

  fclose(fppl1);
  fclose(fppl2);
  fclose(fppl3);
  
  PetscFunctionReturn(0);

}
#undef __FUNCT__
#define __FUNCT__ "ComputeRHS"
PetscErrorCode ComputeRHS(DMMG dmmg, Vec b)
{
  DM             da = dmmg->dm;
  UserContext    *user = (UserContext *) dmmg->user;
  PetscErrorCode ierr;
  PetscInt       i,j,mx,my,xm,ym,mm,nn,xs,ys;
  PetscScalar    Hx,Hy;
  PetscScalar    **array;

  PetscFunctionBegin;

  ierr = DMDAGetInfo(da, 0, &mx, &my, 0, &mm , &nn, 0,0,0,0,0,0,0);CHKERRQ(ierr);
  Hx = dxValue;
  Hy = dyValue;
//  Hx   = 1.0 / (PetscReal)(mx-1);
//  Hy   = 1.0 / (PetscReal)(my-1);
  ierr = DMDAGetCorners(da,&xs,&ys,0,&xm,&ym,0);CHKERRQ(ierr);
  ierr = DMDAVecGetArray(da, b, &array);CHKERRQ(ierr);
//  printf("mx = %d  my = %d  mm = %d  nn = %d  \n", mx, my, mm, nn);
//  printf("xs = %d  ys = %d  xm = %d  ym = %d  \n", xs, ys, xm, ym);
  for (j=ys; j<ys+ym; j++){
    for(i=xs; i<xs+xm; i++){
     // array[j][i] = PetscExpScalar(-((PetscReal)i*Hx)*((PetscReal)i*Hx)/user->nu)*PetscExpScalar(-((PetscReal)j*Hy)*((PetscReal)j*Hy)/user->nu)*Hx*Hy;
     if (i == mx-1) array[j][i] = 0.;
     else if(j >= 0.5*ym && j <= 0.55*ym && i >= 0.5*xm && i <= 0.6*xm) array[j][i] = -j*Hx*Hy;
     else array[j][i] = 0.;

    }
  }
  ierr = DMDAVecRestoreArray(da, b, &array);CHKERRQ(ierr);
  ierr = VecAssemblyBegin(b);CHKERRQ(ierr);
  ierr = VecAssemblyEnd(b);CHKERRQ(ierr);

  /* force right hand side to be consistent for singular matrix */
  /* note this is really a hack, normally the model would provide you with a consistent right handside */
  if (user->bcType == NEUMANN) {
    MatNullSpace nullspace;

    ierr = KSPGetNullSpace(dmmg->ksp,&nullspace);CHKERRQ(ierr);
    ierr = MatNullSpaceRemove(nullspace,b,PETSC_NULL);CHKERRQ(ierr);
  }
  printf("Check\n");
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ComputeRHStest"
PetscErrorCode ComputeRHStest(DMMG dmmg, Vec b)
{
  DM             da = dmmg->dm;
  UserContext    *user = (UserContext *) dmmg->user;
  PetscErrorCode ierr;
  PetscInt       i,j,mx,my,xm,ym,mm,nn,xs,ys;
  PetscScalar    Hx,Hy;
  PetscScalar    **array;

  PetscFunctionBegin;

  ierr = DMDAGetInfo(da, 0, &mx, &my, 0, &mm , &nn, 0,0,0,0,0,0,0);CHKERRQ(ierr);
  Hx = dxValue;
  Hy = dyValue;
//  Hx   = 1.0 / (PetscReal)(mx-1);
//  Hy   = 1.0 / (PetscReal)(my-1);
  ierr = DMDAGetCorners(da,&xs,&ys,0,&xm,&ym,0);CHKERRQ(ierr);
  ierr = DMDAVecGetArray(da, b, &array);CHKERRQ(ierr);
//  printf("mx = %d  my = %d  mm = %d  nn = %d  \n", mx, my, mm, nn);
//  printf("xs = %d  ys = %d  xm = %d  ym = %d  \n", xs, ys, xm, ym);
  for (j=ys; j<ys+ym; j++){
    for(i=xs; i<xs+xm; i++){
     // array[j][i] = PetscExpScalar(-((PetscReal)i*Hx)*((PetscReal)i*Hx)/user->nu)*PetscExpScalar(-((PetscReal)j*Hy)*((PetscReal)j*Hy)/user->nu)*Hx*Hy;
     if (i == mx-1) array[j][i] = 0.;
     else if(j >= 0.5*ym && j <= 0.55*ym && i >= 0.5*xm && i <= 0.6*xm) array[j][i] = -j*Hx*Hy;
     else array[j][i] = 0.;
      
    }
  }
  ierr = DMDAVecRestoreArray(da, b, &array);CHKERRQ(ierr);
  ierr = VecAssemblyBegin(b);CHKERRQ(ierr);
  ierr = VecAssemblyEnd(b);CHKERRQ(ierr);

  /* force right hand side to be consistent for singular matrix */
  /* note this is really a hack, normally the model would provide you with a consistent right handside */
  if (user->bcType == NEUMANN) {
    MatNullSpace nullspace;

    ierr = KSPGetNullSpace(dmmg->ksp,&nullspace);CHKERRQ(ierr);
    ierr = MatNullSpaceRemove(nullspace,b,PETSC_NULL);CHKERRQ(ierr);
  }
  printf("Check_Test\n");
  PetscFunctionReturn(0);
}

    
#undef __FUNCT__
#define __FUNCT__ "ComputeRho"
PetscErrorCode ComputeRho(PetscInt i, PetscInt j, PetscInt mx, PetscInt my, PetscScalar centerRho, PetscScalar *rho)
{
  PetscFunctionBegin;
  if ((i > mx/3.0) && (i < 2.0*mx/3.0) && (j > my/3.0) && (j < 2.0*my/3.0)) {
    *rho = 1.0; 
  } else {
    *rho = 1.0;
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ComputeMatrix"
PetscErrorCode ComputeMatrix(DMMG dmmg, Mat J,Mat jac)
{
  DM             da =  dmmg->dm;
  UserContext    *user = (UserContext *) dmmg->user;
  PetscScalar    centerRho = user->rho;
  PetscErrorCode ierr;
  PetscInt       i,j,mx,my,xm,ym,xs,ys,num;
  PetscScalar    v[5],Hx,Hy,HydHx,HxdHy,rho;
  MatStencil     row, col[5];


  // v[] contains the values of a row of the matrix, which are then set by a call to MatSetValues(). We have multiplied everything by the volume element hx*hy, so if you have 1/hy^2, you now have hx/hy.
  PetscFunctionBegin;
  ierr = DMDAGetInfo(da,0,&mx,&my,0,0,0,0,0,0,0,0,0,0);CHKERRQ(ierr); 
  Hx = dxValue;
  Hy = dyValue; 
//  Hx    = 1.0 / (PetscReal)(mx-1);
//  Hy    = 1.0 / (PetscReal)(my-1);
  HxdHy = Hx/Hy;
  HydHx = Hy/Hx;
  ierr = DMDAGetCorners(da,&xs,&ys,0,&xm,&ym,0);CHKERRQ(ierr);
  for (j=ys; j<ys+ym; j++){
    for(i=xs; i<xs+xm; i++){
      row.i = i; row.j = j;
      ierr = ComputeRho(i, j, mx, my, centerRho, &rho);CHKERRQ(ierr);
      if (i==0 || j==0 || i==mx-1 || j==my-1) {
        if (i==mx-1/*user->bcType == DIRICHLET*/) {
           //v[0] = 2.0*rho*(HxdHy + HydHx);
           v[0] = 1.;
          ierr = MatSetValuesStencil(jac,1,&row,1,&row,v,INSERT_VALUES);CHKERRQ(ierr);
        } else if (i==0 || j==0 || j == my-1 /*user->bcType == NEUMANN*/) {
          num = 0;
          if (j!=0) {
            v[num] = -1.*HxdHy;              col[num].i = i;   col[num].j = j-1;
            num++;
          }
          if (i!=0) {
            v[num] = -1.*HydHx;              col[num].i = i-1; col[num].j = j;
            num++;
          }
          if (i!=mx-1) {
            v[num] = -1.*HydHx;              col[num].i = i+1; col[num].j = j;
            num++;
          }
          if (j!=my-1) {
            v[num] = -1.*HxdHy;              col[num].i = i;   col[num].j = j+1;
            num++;
          }
          v[num]   = (num/2.0)*(HxdHy + HydHx); col[num].i = i;   col[num].j = j;
          num++;
          ierr = MatSetValuesStencil(jac,1,&row,num,col,v,INSERT_VALUES);CHKERRQ(ierr);
        }
      } else {
        v[0] = -1.*HxdHy;              col[0].i = i;   col[0].j = j-1;
        v[1] = -1.*HydHx;              col[1].i = i-1; col[1].j = j;
        v[2] = 2.0*(HxdHy + HydHx); col[2].i = i;   col[2].j = j;
        v[3] = -1.*HydHx;              col[3].i = i+1; col[3].j = j;
        v[4] = -1.*HxdHy;              col[4].i = i;   col[4].j = j+1;
        ierr = MatSetValuesStencil(jac,1,&row,5,col,v,INSERT_VALUES);CHKERRQ(ierr);
      }
    }
  }
  ierr = MatAssemblyBegin(jac,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  ierr = MatAssemblyEnd(jac,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "VecView_VTK"
PetscErrorCode VecView_VTK(Vec x, const char filename[], const char bcName[], DM da)
{

//  MPI_Comm           comm;
//  DM                 da;
//  Vec                coords;
  PetscViewer        viewer;
//  PetscScalar       *array, *values;
//  PetscInt           n, N, maxn, mx, my, dof;
//  PetscInt           i, p;
//  MPI_Status         status;
//  PetscMPIInt        rank, size, tag, nn;
  PetscErrorCode     ierr;
  PetscScalar **array;
  PetscInt i,j, xs, ys, xm, ym;
  FILE *fp;

  fp = fopen("AlanSol.plt", "w");

  PetscFunctionBegin;
  ierr = PetscViewerASCIIOpen(PETSC_COMM_WORLD, filename, &viewer);CHKERRQ(ierr);
  ierr = PetscViewerSetFormat(viewer, PETSC_VIEWER_ASCII_VTK);CHKERRQ(ierr);
  ierr = DMView(da, viewer);CHKERRQ(ierr);
  ierr = VecView(x, viewer);CHKERRQ(ierr);
/*  ierr = VecRestoreArray(x, &array);CHKERRQ(ierr);
  ierr = PetscViewerFlush(viewer);CHKERRQ(ierr);
  ierr = PetscViewerDestroy(&viewer);CHKERRQ(ierr);
  ierr = PetscViewerDestroy(&viewer);CHKERRQ(ierr);*/
  ierr = DMDAGetCorners(da, &xs, &ys, 0, &xm, &ym, 0);
  ierr = DMDAVecGetArray(da, x, &array);
  fprintf(fp, "VARIABLES = \"X\", \"Y\", \"U\"\n");
  fprintf(fp, "ZONE I = %d, J= %d, F=POint\n", xm, ym);
  for(j = ys; j < ys+ym; j++) {
    for(i = xs; i < xs+xm; i++) {
      fprintf(fp, "%f  %f  %f  \n", dxValue*i, dyValue*j, array[j][i]);
    }
  }
  fclose(fp); 
  PetscFunctionReturn(0);
}


